# Copyright 2023 Lawrence Livermore Natioanl Security, LLC
# See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT
from dataclasses import dataclass
from typing import List, Optional


@dataclass
# pylint: disable-next=too-many-instance-attributes
class ContextEntry:
    """
    Represents an entry in the processing queue for directories/files.

    Attributes:
        extractPaths (List[str]): The absolute or relative paths to the files or folders to gather information on.
            Note that Unix style '/' directory separators should be used in paths, even on Windows.
        archive (Optional[str]): The full path, including file name, of the archive file that the files or folders
            in `extractPaths` were extracted from. Used to collect metadata about the overall sample and establish
            "Contains" relationships.
        installPrefix (Optional[str]): The path where the files in `extractPaths` would be if installed
            correctly on an actual system. If not provided, `extractPaths` will be used as the install prefixes.
        omitUnrecognizedTypes (Optional[bool]): If True, files with unrecognized types will be omitted from the generated SBOM.
        includeFileExts (Optional[List[str]]): A list of file extensions to include, even if not recognized by Surfactant.
            `omitUnrecognizedTypes` must be set to True for this to take effect.
        excludeFileExts (Optional[List[str]]): A list of file extensions to exclude, even if recognized by Surfactant.
            If both `omitUnrecognizedTypes` and `includeFileExts` are set, the specified extensions in `includeFileExts`
            will still be included.
        skipProcessingArchive (Optional[bool]): If True, skip processing the given archive file with info extractors.
            Software entry for the archive file will only contain basic information such as hashes. Default is False.
        containerPrefix (Optional[str]): The prefix to use for the generated SBOM's containerPath.  Used to indicate that the
            `extractPaths` specified should map to a specific subfolder within the corresponding archive file.
    """

    extractPaths: List[str]
    archive: Optional[str] = None
    installPrefix: Optional[str] = None
    omitUnrecognizedTypes: Optional[bool] = None
    includeFileExts: Optional[List[str]] = None
    excludeFileExts: Optional[List[str]] = None
    skipProcessingArchive: Optional[bool] = False
    containerPrefix: Optional[str] = None



# Copyright 2025 Lawrence Livermore National Security, LLC
# See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT
# Copyright 2025 Lawrence Livermore National Security, LLC
# See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT
import hashlib
import time
from functools import lru_cache
from pathlib import Path
from typing import Any, Dict, Optional, Union

import requests
import tomlkit
from loguru import logger
from requests.exceptions import RequestException
from tomlkit import parse

from surfactant.configmanager import ConfigManager

RTD_URL = "https://surfactant.readthedocs.io/en/latest/database_sources.toml"


@lru_cache(maxsize=1)
def _get_rtd_raw() -> Optional[str]:
    raw = download_content(RTD_URL)
    logger.debug("Fetched RTD singleton for URL: {}", RTD_URL)
    return raw


def download_content(url: str, timeout: int = 10, retries: int = 3) -> Optional[str]:
    """
    Downloads content from a given URL with retry logic and timeout.

    Args:
        url (str): The URL of the content to download.
        timeout (int): The timeout in seconds for each request attempt.
        retries (int): Number of retry attempts for the download.

    Returns:
        Optional[str]: The content as a string if the request is successful,
                       or None if an error occurs.
    """
    attempt = 0
    while attempt < retries:
        try:
            response = requests.get(url, timeout=timeout)
            if response.status_code == 200:
                logger.debug("Request successful! URL:{}", url)
                return response.text
            if response.status_code == 404:
                logger.debug("Resource not found. URL: {}", url)
                return None
            logger.debug("Unexpected status code {} for URL: {}", response.status_code, url)
        except RequestException as e:
            logger.debug("Attempt {} - Error fetching URL {}: {}", str(attempt + 1), url, e)

        attempt += 1
        sleep_time = 2**attempt  # exponential backoff
        logger.debug("Retrying in {} seconds...", sleep_time)
        time.sleep(sleep_time)

    return None


def calculate_hash(data: str) -> str:
    """
    Calculate the SHA-256 hash of the given data.

    Args:
        data (str): The input string to hash.

    Returns:
        str: The SHA-256 hash of the input string.
    """
    return hashlib.sha256(data.encode("utf-8")).hexdigest()


def _read_toml_file(file_path: Union[str, Path]) -> Optional[Dict[str, Any]]:
    """
    Read and parse a TOML file.

    Args:
        file_path (Union[str, Path]): The path to the TOML file.

    Returns:
        Optional[Dict[str, Any]]: The parsed TOML data, or None if the file does not exist.
    """
    path = Path(file_path) if not isinstance(file_path, Path) else file_path
    try:
        with path.open("r") as f:
            return tomlkit.load(f)
    except FileNotFoundError:
        return None
    except tomlkit.exceptions.TOMLKitError as e:
        raise ValueError(f"Error parsing TOML file at {file_path}: {e}") from e


def _write_toml_file(file_path: Union[str, Path], data: Dict[str, Any]) -> None:
    """
    Write data to a TOML file.

    Args:
        file_path (Union[str, Path]): The path to the TOML file.
        data (Dict[str, Any]): The data to write to the file.
    """
    path = Path(file_path) if not isinstance(file_path, Path) else file_path
    with path.open("w") as f:
        tomlkit.dump(data, f)


def load_db_version_metadata(
    version_file_path: Union[str, Path], database_key: str
) -> Optional[Dict[str, str]]:
    """
    Load the database version metadata for a specific database from the specified TOML file.

    Args:
        version_file_path (Union[str, Path]): The path to the TOML file that tracks database versions.
        database_key (str): The key identifying the database.

    Returns:
        Optional[Dict[str, str]]: A dictionary with metadata for the database, or None if not found.

        Example structure:
        {
            "retirejs": {
                "file": "js_library_patterns_retirejs.json",
                "source": "https://example.com/source.json",
                "hash": "abc123...",
                "timestamp": "2025-02-10T19:18:34.784116Z"
            },
            "abc": {
                "file": "some_other_library_patterns_abc.json",
                "source": "https://example.com/other_source.json",
                "hash": "def456...",
                "timestamp": "2025-02-10T20:00:00.000000Z"
            }
        }
    """
    db_metadata = _read_toml_file(version_file_path)
    if db_metadata is None:
        return None

    return db_metadata.get(database_key, {})


def save_db_version_metadata(version_info: Union[str, Path], database_info: Dict[str, str]) -> None:
    """
    Save the metadata for a specific database to the specified TOML file.

    Args:
        version_info (Union[str, Path]): The path to the TOML file.
        database_info (Dict[str, str]): A dictionary containing keys:
            - "database_key": The key identifying the database  (e.g., "retirejs")..
            - "database_file": The file name of the database (e.g., "js_library_patterns_retirejs.json").
            - "source": The source URL of the database.
            - "hash_value": The hash value of the database file.
            - "timestamp": The timestamp when the database was downloaded.

    Raises:
        ValueError: If required keys are missing from database_info.
    """
    required_keys = {"database_key", "database_file", "source", "hash_value", "timestamp"}
    if not required_keys.issubset(database_info):
        raise ValueError(f"database_info must contain the keys: {required_keys}")

    db_metadata = _read_toml_file(version_info) or {}
    new_data = {
        database_info["database_key"]: {
            "file": database_info["database_file"],
            "source": database_info["source"],
            "hash": database_info["hash_value"],
            "timestamp": database_info["timestamp"],
        }
    }
    db_metadata.update(new_data)
    _write_toml_file(version_info, db_metadata)


def fetch_db_config() -> dict:
    """
    Load the database_sources.toml from local docs/
    """
    # Get local docs copy in the source tree:
    local = Path(__file__).parents[2] / "docs" / "database_sources.toml"
    return _read_toml_file(local)


def get_source_for(database_category: str, key: str) -> str:
    """
    Retrieve the URL for a given database category and key.

    Resolution order:
      1. Runtime override in user config.
      2. Local docs/database_sources.toml when running Surfactant from an editable install from a git clone of the Surfactant repo.
      3. ReadTheDocs hosted database_sources.toml file.
      4. Fallback to hard-coded URL in child class code.
    """

    # First, check the runtime config for an override
    config_manager = ConfigManager()
    config = config_manager.config
    runtime_url = config_manager.get("sources", f"{database_category}.{key}")
    if runtime_url not in ("", [], {}, "[]", "{}", None, "None"):
        logger.debug("Using command line overide url: {}", runtime_url)
        return runtime_url  # Return the command line override URL if present

    # Second, check docs/database_sources.toml for the URL
    config = fetch_db_config()
    try:
        if config:
            logger.debug(
                "Using command docs/database_sources.toml: {}",
                config["sources"][database_category][key],
            )
            return config["sources"][database_category][key]
        logger.debug("Failed to get local docs/database_sources.toml")
    except KeyError:
        logger.debug("No external override found for [{}].{}", database_category, key)

    # Third, check ReadTheDocs
    raw_data = _get_rtd_raw()
    if raw_data:
        config = parse(raw_data)  # Parse the raw TOML text into a tomlkit.document.Document
        try:
            runtime_url = config["sources"][database_category][key]
            logger.debug("Using RTD url: {}", runtime_url)
            return runtime_url  # Return url from RTD
        except tomlkit.exceptions.NonExistentKey:
            logger.debug("No RTD override found for [{}].{}", database_category, key)

    return None  # 4th, fall back to the hardcoded URLs in the code



# Copyright 2025 Lawrence Livermore National Security, LLC
# See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT
# Copyright 2025 Lawrence Livermore National Security, LLC
# See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT
import atexit
import bz2
import gzip
import json
import lzma
import os
import pathlib
import shutil
import tarfile
import tempfile
import zipfile
from queue import Queue
from typing import Any, Callable, Dict, List, Literal, Optional, Tuple, Union

import rarfile
from loguru import logger

import surfactant.plugin
from surfactant import ContextEntry
from surfactant.configmanager import ConfigManager
from surfactant.sbomtypes import SBOM, Software
from surfactant.utils import exit_hook

EXTRACT_DIR = pathlib.Path(
    ConfigManager().get("decompression", "extract_dir", tempfile.gettempdir())
)
EXTRACT_DIRS_PREFIX = ConfigManager().get("decompression", "extract_prefix", "surfactant-temp")
EXTRACT_DIR.mkdir(parents=True, exist_ok=True)

# Global list to track extracted dirs
# Hash -> Path to extracted directory & Result of array of 2-tuples (install_prefix, extract_path)
EXTRACT_DIRS = {}
EXTRACT_DIRS_PATH = EXTRACT_DIR / ".surfactant_extracted_dirs.json"

RAR_SUPPORT = {"enabled": False}


def supports_file(filetype: list[str]) -> Optional[list[str]]:
    if filetype is None:
        return None
    supported_types = {"TAR", "GZIP", "ZIP", "BZIP2", "XZ"}
    supported = []
    # Filter out non-archive types
    for ft in filetype:
        if ft in supported_types:
            supported.append(ft)
        elif ft == "RAR" and RAR_SUPPORT["enabled"]:
            supported.append(ft)
    if supported:
        return supported

    return None


# pylint: disable=too-many-positional-arguments
@surfactant.plugin.hookimpl
def extract_file_info(
    sbom: SBOM,
    software: Software,
    filename: str,
    filetype: List[str],
    context_queue: "Queue[ContextEntry]",
    current_context: Optional[ContextEntry],
) -> Optional[Dict[str, Any]]:
    # Check if the file is compressed and get its format
    compression_format = supports_file(filetype)

    if compression_format:
        for fmt in compression_format:
            create_extraction(
                filename,
                software,
                context_queue,
                current_context,
                lambda f, t, format=fmt: decompress_to(f, t, format),
            )


def create_extraction(
    filename: str,
    software: Software,
    context_queue: "Queue[ContextEntry]",
    current_context: Optional[ContextEntry],
    decompress: Callable[[str, str], Union[bool, List[Tuple[str, str]]]],
):
    """Create extraction context entries for decompressed archive files.

    Args:
        filename (str): Path to the archive file to be extracted
        software (Software): Software object to associated with the file; used to skip extraction if already processed
        context_queue (Queue[ContextEntry]): Queue to add new context entries for extracted content
        current_context (Optional[ContextEntry]): Current context entry being processed
        decompress (Callable[[str, str], Union[bool, List[Tuple[str, str]]]]): Function that performs
            the actual decompression. Takes filename and output folder, returns True/False for success
            or a list of tuples containing (install_prefix, extract_path) pairs for multiple entries
    """

    install_prefix = ""

    # Check that archive key exists and filename is same as archive file
    if current_context and current_context.archive and current_context.archive == filename:
        if current_context.extractPaths is not None and current_context.extractPaths != []:
            logger.info(
                f"Already extracted, skipping extraction for archive: {current_context.archive}"
            )
            return

        # Inherit the context entry install prefix for the extracted files
        install_prefix = current_context.installPrefix

    if (
        software.sha256 in EXTRACT_DIRS
        and EXTRACT_DIRS[software.sha256]["result"]
        and os.path.exists(EXTRACT_DIRS[software.sha256]["path"])
    ):
        entries = EXTRACT_DIRS[software.sha256]["result"]
        logger.info(f"Using cached extraction entries for {filename}")
    else:
        # Create a temporary directory for extraction
        temp_folder = create_extract_dir()
        EXTRACT_DIRS[software.sha256] = {"path": temp_folder, "result": None}

        # Decompress the file
        entries = decompress(filename, temp_folder)

        # Simple case where the decompressor doesn't need multiple entries
        if entries is True:
            entries = [(None, temp_folder)]

        # If False or an empty list
        if not entries:
            logger.error(f"Failed to decompress {filename}. No entries created.")
            return

        # Store the result in the global EXTRACT_DIRS
        EXTRACT_DIRS[software.sha256]["result"] = entries

    for entry_prefix, extract_path in entries:
        # Merges our install prefix with the entry's install prefix (where applicable)
        entry_prefix = "/".join(filter(None, [install_prefix, entry_prefix]))

        # Create a new context entry and add it to the queue
        new_entry = ContextEntry(
            archive=filename,
            installPrefix=entry_prefix,
            extractPaths=[extract_path],
            skipProcessingArchive=True,
        )
        context_queue.put(new_entry)
        logger.info(
            f"New ContextEntry added for extracted files: {extract_path} (prefix: {entry_prefix})"
        )


def decompress_to(filename: str, output_folder: str, compression_format: str) -> bool:
    if compression_format == "ZIP":
        decompress_zip_file(filename, output_folder)
    elif compression_format == "TAR":
        extract_tar_file(filename, output_folder)
    elif compression_format in {"GZIP", "BZIP2", "XZ"}:
        try:
            tar_modes = {
                "GZIP": "r:gz",
                "BZIP2": "r:bz2",
                "XZ": "r:xz",
            }
            extract_tar_file(filename, output_folder, tar_modes[compression_format], True)
        except tarfile.ReadError as e:
            # Check if we expected it to be readable as a compressed tar file
            if (
                ".tar" in pathlib.Path(filename).suffixes
                or ".tgz" in pathlib.Path(filename).suffixes
            ):
                logger.error(f"Error decompressing tar file {filename}: {e}")
                logger.info(
                    f"Attempting to decompress {filename} using the appropriate library as a single file"
                )
            # Since it doesn't seem to be a compressed tar file, try just decompressing the file
            return decompress_file(filename, output_folder, compression_format)
    elif compression_format == "RAR":
        decompress_rar_file(filename, output_folder)
    else:
        raise ValueError(f"Unsupported compression format: {compression_format}")
    return True


def decompress_zip_file(filename: str, output_folder: str):
    try:
        with zipfile.ZipFile(filename, "r") as f:
            f.extractall(path=output_folder)
    except zipfile.BadZipFile as e:
        logger.error(f"Error extracting ZIP file {filename}: {e}")
    logger.info(f"Extracted ZIP contents to {output_folder}")


def decompress_file(
    filename: str, output_folder: str, compression_type: Literal["GZIP", "BZIP2", "XZ"]
) -> bool:
    filepath = pathlib.Path(filename)
    output_filename = filepath.name

    extensions = {
        "GZIP": ".gz",
        "BZIP2": ".bz2",
        "XZ": ".xz",
    }
    if filename.endswith(extensions[compression_type]):
        output_filename = filepath.stem

    modules = {
        "GZIP": gzip,
        "BZIP2": bz2,
        "XZ": lzma,
    }
    try:
        module = modules[compression_type]
        with module.open(filename, "rb") as f_in:
            with open(os.path.join(output_folder, output_filename), "wb") as f_out:
                shutil.copyfileobj(f_in, f_out)
    except gzip.BadGzipFile as e:
        # Likely only the first stream of a concatenated file was decompressed, so we will still keep the temp dir
        logger.warning(
            f"Trailing garbage bytes or concatenated streams ignored for {filename}: {e}"
        )
    except OSError as e:
        logger.warning(f"Unable to decompress {filename}: {e}")
        # This file will be completely unusable.
        return False
    return True


def extract_tar_file(
    filename: str,
    output_folder: str,
    open_mode: Literal["r", "r:*", "r:", "r:gz", "r:bz2", "r:xz"] = "r",
    throw_on_read_error: bool = True,
):
    try:
        with tarfile.open(filename, open_mode) as tar:
            tar.extractall(path=output_folder)
    except FileNotFoundError:
        logger.error(f"File not found: {filename}")
    except tarfile.TarError as e:
        if throw_on_read_error and isinstance(e, tarfile.ReadError):
            raise e
        logger.error(f"Error extracting tar file: {e}")
    logger.info(f"Extracted TAR contents to {output_folder}")


def decompress_rar_file(filename: str, output_folder: str):
    try:
        rf = rarfile.RarFile(filename)
        rf.extractall(path=output_folder)
    except rarfile.Error as e:
        logger.error(f"Error extracting rar file: {e}")
    logger.info(f"Extracted RAR contents to {output_folder}")


def setup_extracted_dirs():
    """Get the list of directories where files have been extracted."""
    should_cache_extractions = ConfigManager().get("decompression", "cache_extractions", True)
    if should_cache_extractions and EXTRACT_DIRS_PATH.exists():
        try:
            with open(EXTRACT_DIRS_PATH, "r") as f:
                GLOBAL_EXTRACT_DIRS = json.load(f)
            if not isinstance(GLOBAL_EXTRACT_DIRS, dict):
                logger.error(f"Invalid format in {EXTRACT_DIRS_PATH}. Expected a dictionary.")
                GLOBAL_EXTRACT_DIRS = {}
        except json.JSONDecodeError:
            logger.error(f"Failed to read extracted directories from {EXTRACT_DIRS_PATH}.")
            GLOBAL_EXTRACT_DIRS = {}
    else:
        GLOBAL_EXTRACT_DIRS = {}


def store_extracted_dirs():
    """Store the current extracted directories to a JSON file."""
    should_cache_extractions = ConfigManager().get("decompression", "cache_extractions", True)
    if should_cache_extractions:
        try:
            with open(EXTRACT_DIRS_PATH, "w") as f:
                json.dump(EXTRACT_DIRS, f)
        except IOError as e:
            logger.error(f"Failed to write extracted directories to {EXTRACT_DIRS_PATH}: {e}")


def create_extract_dir():
    return tempfile.mkdtemp(prefix=EXTRACT_DIRS_PREFIX, dir=EXTRACT_DIR)


def delete_extract_dirs():
    exited_gracefully = exit_hook.has_exited_gracefully()
    should_cache_extractions = ConfigManager().get("decompression", "cache_extractions", True)
    should_persist_extractions = should_cache_extractions and ConfigManager().get(
        "decompression", "persist_extractions", False
    )
    keys = list(EXTRACT_DIRS.keys())
    for key in keys:
        # Extraction was in progress or failed; we have no reason to keep it
        extraction_failed = not EXTRACT_DIRS[key]["result"]
        should_delete = extraction_failed or (not should_persist_extractions and exited_gracefully)
        if not should_cache_extractions or should_delete:
            temp_dir = EXTRACT_DIRS[key]["path"]
            if temp_dir and os.path.exists(temp_dir):
                shutil.rmtree(temp_dir)
                logger.info(f"Cleaned up temporary directory: {temp_dir}")
            del EXTRACT_DIRS[key]


def setup_rar_support():
    RAR_SUPPORT["enabled"] = False

    should_enable_rar = ConfigManager().get("rar", "enabled", True)
    if should_enable_rar:
        try:
            result = rarfile.tool_setup()
            if result.setup["open_cmd"][0] in ("UNRAR_TOOL", "UNAR_TOOL"):
                RAR_SUPPORT["enabled"] = True
                return
        except rarfile.RarCannotExec:
            pass
        logger.warning(
            "Install 'Unrar' or 'unar' tool for RAR archive decompression. RAR decompression disabled until installed."
        )


@surfactant.plugin.hookimpl
def init_hook(command_name: Optional[str] = None):
    """Initialize the file decompression plugin."""
    setup_extracted_dirs()
    setup_rar_support()


@atexit.register
def cleanup_hook():
    """Clean up temporary directories and store extraction cache on exit."""
    delete_extract_dirs()
    store_extracted_dirs()



# Copyright 2025 Lawrence Livermore National Security, LLC
# See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT
# import struct
# from pathlib import Path
# from queue import Queue                       # Present for use in future extraction implementation
from typing import Any, Dict, List, Tuple

import rpmfile
from loguru import logger

import surfactant.plugin
from surfactant.sbomtypes import SBOM, Software

# from surfactant.context import ContextEntry   # Present for use in future extraction implementation


def supports_file(filetype) -> bool:
    """Returns if filetype contains "RPM Package" """
    logger.debug("Checks for RPM Package")
    return "RPM Package" in filetype


def get_files(
    directories: List[bytes], files: List[bytes], indicies: List[int], hashes: List[bytes]
) -> Dict[Any, Any]:
    """Extracts files from the given directories and files list.

    Args:
        directories (List[bytes]): List of directory paths in bytes.
        files (List[bytes]): List of file names in bytes.
        indicies (List[bytes]): Directory associated with current file.
    Returns:
        List of extracted file paths.
    """
    extracted_files = {}
    index = 0
    # If there is only one index, return as an integer, not a list of integers
    if isinstance(indicies, int):
        directory = directories[indicies].decode()
        file_name = files[index].decode()
        file_hash = hashes[index].decode()
        extracted_files[f"{directory}{file_name}"] = file_hash
    else:
        while index < len(indicies):
            directory = directories[indicies[index]].decode()
            file_name = files[index].decode()
            file_hash = hashes[index].decode()
            extracted_files[f"{directory}{file_name}"] = file_hash
            index += 1
    return extracted_files


def combine_lists(key_list: List[bytes], value_list: List[bytes]) -> Dict:
    """Combines 2 lists into a single dictionary - matching on index.

    Args:
        key_list (List[bytes]): List to use as key.
        value_list (List[bytes]): List to use as value.
    Returns:
        Combined dictionary.
    """
    result = {}
    if not key_list or not value_list or len(key_list) != len(value_list):
        return result
    index = 0
    while index < len(key_list):
        name = key_list[index].decode()
        version = value_list[index].decode()
        result[name] = version
        index += 1
    return result


@surfactant.plugin.hookimpl
def extract_file_info(
    sbom: SBOM,
    software: Software,
    filename: str,
    filetype: List[str],
    software_field_hints: List[Tuple[str, object, int]],
    # context_queue: "Queue[ContextEntry]",     # Present for use in future extraction implementation
    # current_context: Optional[ContextEntry],
) -> object:
    if not supports_file(filetype):
        return None
    rpm_info = extract_rpm_info(filename)

    for key in rpm_info["rpm"]:
        software_field_hints.append((key, rpm_info["rpm"][key], 80))
    return rpm_info


def extract_rpm_info(filename: str) -> Dict[str, Any]:
    """Extracts fields from the header of an RPM Package.

    Args:
        filename: Path to file to extract information from.
    Returns:
        A dictionary of all fields found in the RPM header.
    """
    file_details: Dict[str, Any] = {}
    with rpmfile.open(filename) as rpm:
        header = rpm.headers
        file_details["rpm"] = {}
        # If any additional fields are desired that have values of single strings, just add the field to this list
        simple_keys = [
            "name",
            "sourcerpm",
            "version",
            "release",
            "summary",
            "description",
            "rpmversion",
            "copyright",
            "os",
            "arch",
            "target",
            "url",
            "archive_format",
            "archive_compression",
            "optflags",
            "sha256",
            "md5",
        ]
        for key in simple_keys:
            if key in header:
                file_details["rpm"][key] = header[key].decode()
        # Handle sections that aren't simple strings
        complex_keys = {
            "requirename": "requireversion",
            "conflict": "conflictversion",
            "obsoletes": "obsoleteversion",
            "enhancename": "enhanceversion",
            "suggestname": "suggestversion",
            "recommendname": "recommendversion",
            "supplementname": "supplementversion",
            "provides": "provideversion",
        }
        for key, value in complex_keys.items():
            if key in header:
                file_details["rpm"][key] = combine_lists(header[key], header[value])
        if "buildtime" in header:
            file_details["rpm"]["buildtime"] = header["buildtime"]
        if "basenames" in header:
            file_details["rpm"]["associated_files"] = get_files(
                header["dirnames"], header["basenames"], header["dirindexes"], header["filemd5s"]
            )
        return file_details



import json
import subprocess
import sys
import uuid
from datetime import datetime, timezone
from pathlib import Path
from typing import List

from loguru import logger

import surfactant.plugin
from surfactant.sbomtypes import SBOM, Software


def run_cve_bin_tool(input_file_path, shaHash, output_dir):
    cvebin_file_name = f"{shaHash}_{input_file_path.stem}.json"
    output_file_path = output_dir / cvebin_file_name

    cdxvex_file_name = f"{shaHash}_{input_file_path.stem}.cdxvex"
    vex_output_path = output_dir / cdxvex_file_name

    try:
        command = [
            "cve-bin-tool",
            "--offline",
            "--input-file",
            str(input_file_path),
            "--output",
            str(output_file_path),
            "--format",
            "json",
            "--vex",
            str(vex_output_path),
        ]
        result = subprocess.run(command, capture_output=True, text=True, check=False)

        # Check the exit status
        if result.returncode in (0, 1):
            return output_file_path  # Return path to the generated JSON file
    except subprocess.CalledProcessError as e:
        logger.error(
            f"Error running CVE-bin-tool: {e}\nOutput: {e.output}\nError: {e.stderr}",
            file=sys.stderr,
        )
    return None


def convert_cve_to_openvex(json_output_path, shaHash, output_dir):
    openvex_file_name = f"{json_output_path.stem}.vex"
    openvex_output = output_dir / openvex_file_name

    # Open and read the .json file
    try:
        with open(json_output_path, "r") as file:
            cve_data = json.load(file)
    except json.JSONDecodeError as e:
        logger.error(f"Error reading JSON file: {e}")
        return
    except IOError as e:
        logger.error(f"IO error when reading {json_output_path}: {e}")

    openvex_template = {
        "@context": "https://openvex.dev/ns/v0.2.0",
        "@id": f"urn:uuid:{uuid.uuid4()}",
        "author": "Surfactant plugin cvebintool2vex",
        "timestamp": datetime.now(timezone.utc).isoformat(timespec="microseconds"),
        "version": 1,
        "tooling": "Surfactant (https://github.com/LLNL/Surfactant)",
        "statements": [],
    }

    for entry in cve_data:
        # Convert CVE data to OpenVEX format
        statement = {
            "vulnerability": {"name": entry["cve_number"]},
            "products": [
                {"@id": f"cpe:2.3:a:{entry['vendor']}:{entry['product']}:{entry['version']}:::::"}
            ],
            "status": "under_investigation",
            "source": entry["source"],
            "cvss_version": entry["cvss_version"],
            "cvss_vector": entry["cvss_vector"],
            "severity": entry["severity"],
        }
        openvex_template["statements"].append(statement)

    # Save the OpenVEX output to a new file
    try:
        with open(openvex_output, "w") as outfile:
            json.dump(openvex_template, outfile, indent=4)
    except IOError as e:
        logger.error(f"IO error when writing {openvex_output}: {e}")


def process_input(input_path, shaHash, output_dir=None):
    input_path = Path(input_path)
    if output_dir is None:
        output_dir = Path.cwd()

    if input_path.is_dir():
        for input_file in input_path.glob("*.*"):
            if input_file.suffix.lower() not in [".bin", ".exe", ".jar"]:
                continue
            process_file(input_file, shaHash, output_dir)
    elif input_path.is_file():
        process_file(input_path, shaHash, output_dir)
    else:
        logger.info(f"Error: {input_path} is neither a file nor a directory.")


def process_file(input_file, shaHash, output_directory):
    try:
        json_output_path = run_cve_bin_tool(input_file, shaHash, output_directory)
        if json_output_path:
            convert_cve_to_openvex(json_output_path, shaHash, output_directory)
    except subprocess.CalledProcessError as e:
        logger.error(f"Error running CVE-bin-tool: {e}")
    except json.JSONDecodeError as e:
        logger.error(f"JSON decoding error in {input_file}: {e}")
    except IOError as e:
        logger.error(f"I/O error processing {input_file}: {e}")

    # Check if the expected JSON file was created and proceed if it exists
    cvebin_file_name = f"{shaHash}_{input_file.stem}.json"
    jsonfile = output_directory / cvebin_file_name
    if not jsonfile.exists():
        logger.warning(f"Expected JSON file does not exist: {jsonfile}")


def delete_extra_files(*file_paths):
    for file_path in file_paths:
        try:
            if file_path.exists():
                file_path.unlink()
        except PermissionError as e:
            logger.error(f"Permission error deleting {file_path}: {e}")

        except OSError as e:
            logger.error(f"OS error deleting {file_path}: {e}")


@surfactant.plugin.hookimpl(specname="extract_file_info")
def cvebintool2vex(sbom: SBOM, software: Software, filename: str, filetype: List[str]):
    """
    :param sbom(SBOM): The SBOM that the software entry/file is being added to. Can be used to add observations or analysis data.
    :param software(Software): The software entry associated with the file to extract information from.
    :param filename (str): The full path to the file to extract information from.
    :param filetype (List[str]): File type information based on magic bytes.
    """
    # Only parsing executable files
    if not ("ELF" in filetype or "PE" in filetype):
        return

    shaHash = str(software.sha256)
    filename = Path(filename)
    output_dir = Path.cwd()

    existing_json_path = output_dir / f"{shaHash}_additional_metadata.json"
    if existing_json_path.exists():
        with open(existing_json_path, "r") as file:
            data = json.load(file)
    else:
        data = {
            "sha256hash": shaHash,
            "filename": [filename.name],
            "openvex": [],
            "cyclonedx-vex": [],
            "cve-bin-tool": [],
        }

    # Assuming JSON, CDXVEX, and VEX files are processed here
    process_input(filename, shaHash, output_dir)
    # and you have the output files: .json, .cdxvex, .vex

    # Integrate .cdxvex and .vex file contents
    cdxvex_file_path = output_dir / f"{shaHash}_{filename.stem}.cdxvex"
    vex_file_path = output_dir / f"{shaHash}_{filename.stem}.vex"
    json_file_path = output_dir / f"{shaHash}_{filename.stem}.json"

    if cdxvex_file_path.exists() and vex_file_path.exists() and json_file_path.exists():
        # For .cdxvex and .vex files, if they contain JSON, parse them as such; otherwise, read as text
        try:
            with open(cdxvex_file_path, "r") as file:
                cdxvex_data = json.load(file)  # Assuming .cdxvex file is in JSON format
            data["cyclonedx-vex"].append(cdxvex_data)
        except json.JSONDecodeError:
            with open(cdxvex_file_path, "r") as file:
                cdxvex_data = file.read()  # Fallback if not JSON
            data["cyclonedx-vex"].append(cdxvex_data)

        try:
            with open(vex_file_path, "r") as file:
                vex_data = json.load(file)  # Assuming .vex file is in JSON format
            data["openvex"].append(vex_data)
        except json.JSONDecodeError:
            with open(vex_file_path, "r") as file:
                vex_data = file.read()  # Fallback if not JSON
            data["openvex"].append(vex_data)

        with open(json_file_path, "r") as file:
            json_data = json.load(file)
        data["cve-bin-tool"].append(json_data)

    # Attempt to save the updated data
    try:
        with open(existing_json_path, "w") as file:
            json.dump(data, file, indent=4)
            logger.info(f"Updated data saved to {existing_json_path}")
    except IOError as e:
        logger.error(f"IO error when writing {existing_json_path}: {e}")

    # Clean up extra files
    delete_extra_files(cdxvex_file_path, vex_file_path, json_file_path)


@surfactant.plugin.hookimpl
def update_db():
    # Example update logic
    try:
        subprocess.check_call(["cve-bin-tool", "--update", "now", "."])
        return "Database updated successfully."
    except subprocess.CalledProcessError as e:
        return f"Failed to update database: {e}"



import sys
from typing import Any, Callable, Optional


# https://stackoverflow.com/a/9741784
class ExitHooks:
    def __init__(self):
        self.exit_code: Optional[int] = None
        self.exception: Optional[Exception] = None
        self._orig_exit: Optional[Callable[[int], None]] = None

    def hook(self):
        """Install the exit and exception hooks."""
        self._orig_exit = sys.exit
        sys.exit = self.exit
        sys.excepthook = self.exc_handler

    def exit(self, code: int = 0):
        """Custom exit handler that captures the exit code."""
        self.exit_code = code
        if self._orig_exit is not None:
            self._orig_exit(code)

    def exc_handler(self, exc_type: type, exc: Exception, *args: Any):
        """Custom exception handler that captures exceptions."""
        self.exception = exc


_HOOKS = ExitHooks()
_HOOKS.hook()


def get_exit_code() -> Optional[int]:
    """Get the exit code from the last program exit."""
    return _HOOKS.exit_code


def get_exception() -> Optional[Exception]:
    """Get the exception from the last unhandled exception."""
    return _HOOKS.exception


def has_exited_gracefully() -> bool:
    """
    Returns True if the program exited gracefully (without an exception).
    """
    if _HOOKS.exit_code is not None and _HOOKS.exit_code != 0:
        return False
    return _HOOKS.exception is None



# Copyright 2023 Lawrence Livermore National Security, LLC
# See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT
from __future__ import annotations

import json
import uuid as uuid_module
from dataclasses import asdict, dataclass, field, fields
from typing import Dict, List, Optional, Set

import networkx as nx
from dataclasses_json import config, dataclass_json
from loguru import logger

from ._analysisdata import AnalysisData
from ._file import File
from ._hardware import Hardware
from ._observation import Observation
from ._provenance import SoftwareProvenance
from ._relationship import Relationship, StarRelationship
from ._software import Software, SoftwareComponent
from ._system import System

INTERNAL_FIELDS = {"software_lookup_by_sha256"}


def recover_serializers(cls):
    """
    After dataclass_json has bound its own to_dict/to_json,
    restore any _to_dict/_to_json overrides.
    """
    if hasattr(cls, "to_dict_override"):
        cls.to_dict = cls.to_dict_override
    if hasattr(cls, "to_json_override"):
        cls.to_json = cls.to_json_override
    return cls


@recover_serializers
@dataclass_json
@dataclass
class SBOM:
    # pylint: disable=R0902
    systems: List[System] = field(default_factory=list)
    hardware: List[Hardware] = field(default_factory=list)
    software: List[Software] = field(default_factory=list)
    # relationships: Set[Relationship] = field(default_factory=set)  # (removed relationships field. Graph is now the single source of truth)
    _loaded_relationships: List[Relationship] = (
        field(  # this metadata will capture the old array on load (but won’t re-emit it)
            default_factory=list,
            metadata=config(field_name="relationships", exclude=lambda _: True),
        )
    )
    analysisData: List[AnalysisData] = field(default_factory=list)
    observations: List[Observation] = field(default_factory=list)
    starRelationships: Set[StarRelationship] = field(default_factory=set)
    software_lookup_by_sha256: Dict = field(default_factory=dict)
    graph: nx.MultiDiGraph = field(
        init=False,
        repr=False,
        # metadata=config(exclude=lambda _: True),  # internal graph; excluded from JSON
        metadata=config(exclude=lambda _: True),
    )  # Add a NetworkX directed graph for quick traversal/query

    def __post_init__(self):
        # If called like SBOM(raw_dict), raw_dict will be in .systems
        if isinstance(self.systems, dict) and not self.hardware and not self.software:
            raw = self.systems

            # zero out every container
            self.systems = []
            self.hardware = []
            self.software = []
            self.analysisData = []
            self.observations = []
            self._loaded_relationships = []
            self.starRelationships = set()

            # prepare valid field-name sets
            SYSTEM_FIELDS = {f.name for f in fields(System)}
            HARDWARE_FIELDS = {f.name for f in fields(Hardware)}
            SOFTWARE_FIELDS = {f.name for f in fields(Software)}
            REL_FIELDS = {f.name for f in fields(Relationship)}
            AD_FIELDS = {f.name for f in fields(AnalysisData)}
            OBS_FIELDS = {f.name for f in fields(Observation)}
            STAR_FIELDS = {f.name for f in fields(StarRelationship)}

            # rehydrate systems
            for sys_data in raw.get("systems", []):
                clean = {k: v for k, v in sys_data.items() if k in SYSTEM_FIELDS}
                self.systems.append(System(**clean))

            # rehydrate hardware
            for hw_data in raw.get("hardware", []):
                clean = {k: v for k, v in hw_data.items() if k in HARDWARE_FIELDS}
                self.hardware.append(Hardware(**clean))

            # rehydrate software
            for sw_data in raw.get("software", []):
                clean = {k: v for k, v in sw_data.items() if k in SOFTWARE_FIELDS}
                self.software.append(Software(**clean))

            # rehydrate relationships into the loader list
            for rel_data in raw.get("relationships", []):
                clean = {k: v for k, v in rel_data.items() if k in REL_FIELDS}
                self._loaded_relationships.append(Relationship(**clean))

            # rehydrate analysisData
            for ad_data in raw.get("analysisData", []):
                clean = {k: v for k, v in ad_data.items() if k in AD_FIELDS}
                self.analysisData.append(AnalysisData(**clean))

            # rehydrate observations
            for obs_data in raw.get("observations", []):
                clean = {k: v for k, v in obs_data.items() if k in OBS_FIELDS}
                self.observations.append(Observation(**clean))

            # rehydrate starRelationships
            for sr_data in raw.get("starRelationships", []):
                clean = {k: v for k, v in sr_data.items() if k in STAR_FIELDS}
                self.starRelationships.add(StarRelationship(**clean))

        # Strip out internal-only fields so dataclass logic and JSON serializers ignore them
        # pylint: disable=access-member-before-definition
        self.__dataclass_fields__ = {
            k: v for k, v in self.__dataclass_fields__.items() if k not in INTERNAL_FIELDS
        }

        # Build the NetworkX graph from systems/software and loaded relationships
        self.build_graph()

    def build_graph(self) -> None:
        """Rebuild the directed graph from systems, software, and any loaded relationships."""
        self.graph = nx.MultiDiGraph()
        for sys in self.systems:
            self.graph.add_node(sys.UUID, type="System")
        for sw in self.software:
            self.graph.add_node(sw.UUID, type="Software")
        # rehydrate edges from loaded JSON (if any)
        for rel in self._loaded_relationships:
            self.graph.add_edge(rel.xUUID, rel.yUUID, key=rel.relationship)

    def add_relationship(self, rel: Relationship) -> None:
        # The Relationship object get wired into the graph key=…
        if not self.graph.has_node(rel.xUUID):
            self.graph.add_node(rel.xUUID, type="Unknown")
        if not self.graph.has_node(rel.yUUID):
            self.graph.add_node(rel.yUUID, type="Unknown")

        # the edge key is the relationship type
        self.graph.add_edge(rel.xUUID, rel.yUUID, key=rel.relationship)

    def create_relationship(self, xUUID: str, yUUID: str, relationship: str) -> Relationship:
        # ensure nodes exist
        if not self.graph.has_node(xUUID):
            self.graph.add_node(xUUID, type="Unknown")
        if not self.graph.has_node(yUUID):
            self.graph.add_node(yUUID, type="Unknown")

        # record the edge, keyed by the relationship
        self.graph.add_edge(xUUID, yUUID, key=relationship)

        # return a Relationship object for backwards-compat
        return Relationship(xUUID, yUUID, relationship)

    def find_relationship_object(self, r: Relationship) -> bool:
        """
        Return True if an exact edge (r.xUUID → r.yUUID) exists
        in the graph with key=r.relationship.
        """
        return self.graph.has_edge(r.xUUID, r.yUUID, key=r.relationship)

    def find_relationship(self, xUUID: str, yUUID: str, relationship: str) -> bool:
        """
        Same as find_relationship_object, but takes raw args.
        """
        return self.graph.has_edge(xUUID, yUUID, key=relationship)

    def has_relationship(
        self,
        xUUID: Optional[str] = None,
        yUUID: Optional[str] = None,
        relationship: Optional[str] = None,
    ) -> bool:
        """
        Return True if there exists at least one edge u→v in the graph
        matching the optional filters (u, v, and/or relationship key).
        """
        # Fast-path if all three are specified
        if xUUID and yUUID and relationship:
            return self.graph.has_edge(xUUID, yUUID, key=relationship)

        # Otherwise scan all edges with keys
        for u, v, key in self.graph.edges(keys=True):
            if xUUID and u != xUUID:
                continue
            if yUUID and v != yUUID:
                continue
            if relationship and key.upper() != relationship.upper():
                continue
            return True
        return False

    def find_software(self, sha256: Optional[str]) -> Optional[Software]:
        if sha256 in self.software_lookup_by_sha256:
            return self.software_lookup_by_sha256[sha256]
        return None

    def add_software(self, sw: Software) -> None:
        if sw.sha256 is not None:
            self.software_lookup_by_sha256[sw.sha256] = sw
        self.software.append(sw)

        # Add a node for the new software
        if not self.graph.has_node(sw.UUID):
            self.graph.add_node(sw.UUID, type="Software")

    def add_software_entries(
        self, entries: Optional[List[Software]], parent_entry: Optional[Software] = None
    ):
        """Add software entries, merging duplicates and preserving all relationship edges.

        Args:
            entries (Optional[List[Software]]): A list of Software entries to add to the SBOM.
            parent_entry (Optional[Software]): An optional parent software entry to add "Contains" relationships to.
        """
        if not entries:
            return
        # if a software entry already exists with a matching file hash, augment the info in the existing entry
        for e in entries:
            existing = self.find_software(e.sha256)
            if existing and Software.check_for_hash_collision(existing, e):
                logger.warning(f"Hash collision between {existing.name} and {e.name}")

            if not existing:
                # new software → add node
                self.add_software(e)
                entry_uuid = e.UUID
            else:
                # duplicate → merge and redirect edges
                kept_uuid, old_uuid = existing.merge(e)

                # redirect *incoming* edges to the kept node
                for src, _, key, attrs in list(self.graph.in_edges(old_uuid, keys=True, data=True)):
                    self.graph.add_edge(src, kept_uuid, key=key, **attrs)

                # redirect *outgoing* edges from the old node
                for _, dst, key, attrs in list(
                    self.graph.out_edges(old_uuid, keys=True, data=True)
                ):
                    self.graph.add_edge(kept_uuid, dst, key=key, **attrs)

                # remove the old UUID entirely
                if self.graph.has_node(old_uuid):
                    self.graph.remove_node(old_uuid)
                entry_uuid = kept_uuid

            # if a parent/package container was provided, attach a "Contains" edge
            if parent_entry:
                parent_uuid = parent_entry.UUID
                if not self.graph.has_edge(parent_uuid, entry_uuid, key="Contains"):
                    self.graph.add_edge(parent_uuid, entry_uuid, key="Contains")

    # pylint: disable=too-many-arguments
    def create_software(
        self,
        *,  # all arguments are keyword-only
        name: Optional[str] = None,
        size: Optional[int] = None,
        sha1: Optional[str] = None,
        sha256: Optional[str] = None,
        md5: Optional[str] = None,
        fileName: Optional[List[str]] = None,
        installPath: Optional[List[str]] = None,
        containerPath: Optional[List[str]] = None,
        captureTime: Optional[int] = None,
        version: Optional[str] = None,
        vendor: Optional[List[str]] = None,
        description: Optional[str] = None,
        relationshipAssertion: Optional[str] = None,
        comments: Optional[str] = None,
        metadata: Optional[List[object]] = None,
        supplementaryFiles: Optional[List[File]] = None,
        provenance: Optional[List[SoftwareProvenance]] = None,
        recordedInstitution: Optional[str] = None,
        components: Optional[List[SoftwareComponent]] = None,
    ) -> Software:
        sw = Software(
            name=name,
            size=size,
            sha1=sha1,
            sha256=sha256,
            md5=md5,
            fileName=fileName,
            installPath=installPath,
            containerPath=containerPath,
            captureTime=captureTime,
            version=version,
            vendor=vendor,
            description=description,
            relationshipAssertion=relationshipAssertion,
            comments=comments,
            metadata=metadata,
            supplementaryFiles=supplementaryFiles,
            provenance=provenance,
            recordedInstitution=recordedInstitution,
            components=components,
        )
        self.software_lookup_by_sha256[sw.sha256] = sw
        self.software.append(sw)
        return sw

    def merge(self, sbom_m: SBOM):
        # merged/old to new UUID map
        uuid_updates: Dict[str, str] = {}

        # 1) Merge systems
        if sbom_m.systems:
            for system in sbom_m.systems:
                # check for duplicate UUID/name, merge with existing entry
                if existing_system := self._find_systems_entry(uuid=system.UUID, name=system.name):
                    # merge system entries
                    u1, u2 = existing_system.merge(system)
                    logger.info(f"MERGE_DUPLICATE_SYS: uuid1={u1}, uuid2={u2}")
                    uuid_updates[u2] = u1

                    # Redirect any existing edges from u2 -> u1 in self.graph
                    if hasattr(self, "graph") and self.graph.has_node(u2):
                        # for each predecessor of u2, add edge (pred -> u1)
                        # Redirect incoming edges to the merged node u1
                        for pred, _, key, attrs in self.graph.in_edges(u2, keys=True, data=True):
                            self.graph.add_edge(pred, u1, key=key, **attrs)

                        # For each successor of u2, add edge (u1 -> succ)
                        # Redirect outgoing edges from u2 → u1
                        for _, succ, key, attrs in self.graph.out_edges(u2, keys=True, data=True):
                            self.graph.add_edge(u1, succ, key=key, **attrs)

                        # Then drop the old node
                        self.graph.remove_node(u2)

                else:
                    self.systems.append(system)

                    # Add the new system node into the graph
                    if hasattr(self, "graph"):
                        self.graph.add_node(system.UUID, type="System")

        # 2) Merge software
        if sbom_m.software:
            for sw in sbom_m.software:
                # NOTE: Do we want to pass in teh UUID here? What if we have two different UUIDs for the same file? Should hashes be required?
                if existing_sw := self._find_software_entry(
                    uuid=sw.UUID, sha256=sw.sha256, md5=sw.md5, sha1=sw.sha1
                ):
                    u1, u2 = existing_sw.merge(sw)
                    logger.info(f"MERGE DUPLICATE: uuid1={u1}, uuid2={u2}")
                    uuid_updates[u2] = u1

                    # Redirect any existing edges from u2 -> u1 in self.graph
                    if hasattr(self, "graph") and self.graph.has_node(u2):
                        # Redirect incoming edges to the merged node u1
                        for pred, _, key, attrs in self.graph.in_edges(u2, keys=True, data=True):
                            self.graph.add_edge(pred, u1, key=key, **attrs)
                        # Redirect outgoing edges from u2 → u1
                        for _, succ, key, attrs in self.graph.out_edges(u2, keys=True, data=True):
                            self.graph.add_edge(u1, succ, key=key, **attrs)

                        # Then drop the old node
                        self.graph.remove_node(u2)

                else:
                    self.software.append(sw)

                    # Add the new software node in the graph
                    if hasattr(self, "graph"):
                        self.graph.add_node(sw.UUID, type="Software")

        # 3) Merge relationships from the incoming SBOM’s MultiDiGraph
        for src, dst, rel_type in sbom_m.graph.edges(keys=True):
            # apply any UUID remaps from merged systems/software
            xUUID = uuid_updates.get(src, src)
            yUUID = uuid_updates.get(dst, dst)

            # skip exact duplicates
            if self.graph.has_edge(xUUID, yUUID, key=rel_type):
                logger.info(f"DUPLICATE RELATIONSHIP: {xUUID} → {yUUID} [{rel_type}]")
            else:
                # add a new edge, keyed by the relationship
                self.graph.add_edge(xUUID, yUUID, key=rel_type)

        # 4) Rewrite any containerPath UUIDs
        for sw in self.software:
            if sw.containerPath:
                for idx, path in enumerate(sw.containerPath):
                    u = path[:36]
                    # if container path starts with an invalid uuid4, sbom might not be valid
                    if self.is_valid_uuid4(u):
                        if u in uuid_updates:
                            updated_path = path.replace(u, uuid_updates[u], 1)
                            sw.containerPath[idx] = updated_path
                # remove duplicates
                sw.containerPath = [*set(sw.containerPath)]

        logger.info(f"UUID UPDATES: {uuid_updates}")

        # 5) Merge analysisData, observations, starRelationships
        for ad in sbom_m.analysisData:
            self.analysisData.append(ad)
        for obs in sbom_m.observations:
            self.observations.append(obs)
        for star_rel in sbom_m.starRelationships:
            # rewrite UUIDs before doing the search
            if star_rel.xUUID in uuid_updates:
                star_rel.xUUID = uuid_updates[star_rel.xUUID]
            if star_rel.yUUID in uuid_updates:
                star_rel.yUUID = uuid_updates[star_rel.yUUID]
            if existing_star_rel := self._find_star_relationship_entry(
                xUUID=star_rel.xUUID,
                yUUID=star_rel.yUUID,
                relationship=star_rel.relationship,
            ):
                logger.info(f"DUPLICATE STAR RELATIONSHIP: {existing_star_rel}")
            else:
                self.starRelationships.add(star_rel)

    def _find_systems_entry(
        self, uuid: Optional[str] = None, name: Optional[str] = None
    ) -> Optional[System]:
        """Merge helper function to find and return
        the matching system entry in the provided sbom.

        Args:
            uuid (Optional[str]): The uuid of the desired system entry.
            name (Optional[str]): The name of the desired system entry.

        Returns:
            Optional[System]: The system found that matches the given criteria, otherwise None.
        """
        for system in self.systems:
            if uuid:
                if system.UUID != uuid:
                    continue
            if name:
                if system.name != name:
                    continue
            return system
        return None

    def _find_software_entry(
        self,
        uuid: Optional[str] = None,
        sha256: Optional[str] = None,
        md5: Optional[str] = None,
        sha1: Optional[str] = None,
    ) -> Optional[Software]:
        """Merge helper function to find and return
        the matching software entry in the provided sbom.

        Args:
            uuid (Optional[str]): The uuid of the desired software entry to match against if no hashes were provided.
            sha256 (Optional[str]): The sha256 of the desired software entry.
            md5 (Optional[str]): The md5 of the desired software entry.
            sha1 (Optional[str]): The sha1 of the desired software entry.

        Returns:
            Optional[Software]: The software entry found that matches the given criteria, otherwise None.
        """
        for sw in self.software:
            match = False
            # If we have hashes to check
            if sha256 or md5 or sha1:
                # Check if we have both sides of the comparison, then compare. At least one hash must match
                if sw.sha256 and sha256:
                    if sw.sha256 == sha256:
                        match = True
                if sw.md5 and md5:
                    if sw.md5 == md5:
                        match = True
                if sw.sha1 and sha1:
                    if sw.sha1 == sha1:
                        match = True
            # If no hashes to check, match by UUID
            else:
                if sw.UUID == uuid:
                    match = True
            if match:
                return sw
        return None

    def _find_relationship_entry(
        self,
        xUUID: Optional[str] = None,
        yUUID: Optional[str] = None,
        relationship: Optional[str] = None,
    ) -> Optional[Relationship]:
        """Merge helper function to find and return
        the matching relationship entry in the provided sbom.

        Locate and return the first matching Relationship object
        found in the MultiDiGraph.

        Args:
            xUUID (Optional[str]): The xUUID of the desired relationship entry.
            yUUID (Optional[str]): The yUUID of the desired relationship entry.
            relationship (Optional[str]): The relationship type of the desired relationship entry.

        Returns:
            Optional[Relationship]: The relationship entry found that matches the given criteria, otherwise None.
        """
        for u, v, key in self.graph.edges(keys=True):
            if xUUID and u != xUUID:
                continue
            if yUUID and v != yUUID:
                continue
            if relationship and key.upper() != relationship.upper():
                continue
            # reconstruct the Relationship object for merge‐logic
            return Relationship(xUUID=u, yUUID=v, relationship=key)
        return None

    def _find_star_relationship_entry(
        self,
        xUUID: Optional[str] = None,
        yUUID: Optional[str] = None,
        relationship: Optional[str] = None,
    ) -> Optional[StarRelationship]:
        """Merge helper function to find and return
        the matching star relationship entry in the provided sbom.

        Args:
            xUUID (Optional[str]): The xUUID of the desired relationship entry.
            yUUID (Optional[str]): The yUUID of the desired relationship entry.
            relationship (Optional[str]): The relationship type of the desired relationship entry.

        Returns:
            Optional[StarRelationship]: The star relationship found that matches the given criteria, otherwise None.
        """
        for rel in self.starRelationships:
            if xUUID:
                if rel.xUUID != xUUID:
                    continue
            if yUUID:
                if rel.yUUID != yUUID:
                    continue
            if relationship:
                if rel.relationship != relationship:
                    continue
            return rel
        return None

    def is_valid_uuid4(self, u: str) -> bool:
        """Merge helper function to check if a uuid is valid.

        Args:
            u (str):  The UUID to check.

        Returns:
            bool: True if the UUID is valid, otherwise False.
        """
        try:
            u_test = uuid_module.UUID(u, version=4)
        except ValueError:
            return False
        return str(u_test) == u

    def get_children(self, xUUID: str, rel_type: Optional[str] = None) -> List[str]:
        """
        Return all v such that there is an edge xUUID → v,
        optionally filtered by relationship key.
        """
        children = []
        for _, v, key in self.graph.out_edges(xUUID, keys=True):
            if rel_type is None or key.upper() == rel_type.upper():
                children.append(v)
        return children

    def get_parents(self, yUUID: str, rel_type: Optional[str] = None) -> List[str]:
        """
        Return all u such that there is an edge u → yUUID,
        optionally filtered by relationship key.
        """
        parents = []
        for u, _, key in self.graph.in_edges(yUUID, keys=True):
            if rel_type is None or key.upper() == rel_type.upper():
                parents.append(u)
        return parents

    def to_dict_override(self) -> dict:
        """
        Dump all SBOM dataclass fields (via asdict), strip out internal-only
        fields, convert sets→lists, and then build a fresh
        'relationships' list by iterating every edge key in the MultiDiGraph.
        """
        # Grab everything as a dict
        data = asdict(self)

        # Remove fields we never want in JSON
        data.pop("graph", None)
        data.pop("_loaded_relationships", None)

        # Turn any sets into lists for JSON
        for k, v in list(data.items()):
            if isinstance(v, set):
                data[k] = list(v)

        # Rebuild 'relationships' from the graph's edge keys
        data["relationships"] = [
            {"xUUID": u, "yUUID": v, "relationship": key}
            for u, v, key in self.graph.edges(keys=True)
        ]

        return data

    def to_json_override(self, *args, **kwargs) -> str:
        """
        Serialize via our to_dict_override, passing through any json.dumps kwargs.
        """
        return json.dumps(self.to_dict_override(), *args, **kwargs)



# Copyright 2025 Lawrence Livermore National Security, LLC
# See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT
from concurrent.futures import ThreadPoolExecutor, as_completed
from pathlib import Path
from queue import Queue
from typing import Any, Dict, List, Optional, Tuple, Union

import olefile
import pymsi
from loguru import logger
from pymsi.msi.component import Component
from pymsi.msi.directory import Directory
from pymsi.thirdparty.refinery.cab import CabFolder

import surfactant.plugin
from surfactant.configmanager import ConfigManager
from surfactant.context import ContextEntry
from surfactant.infoextractors import file_decompression
from surfactant.sbomtypes import SBOM, Software

# https://learn.microsoft.com/en-us/windows/win32/msi/property-reference#system-folder-properties
DEFAULT_PATHS = {
    "AdminToolsFolder": "C:\\ProgramData\\Microsoft\\Windows\\Start Menu\\Programs\\Administrative Tools",
    "AppDataFolder": "C:\\Users\\USER\\AppData\\Roaming",
    "CommonAppDataFolder": "C:\\ProgramData",
    "CommonFiles64Folder": "C:\\Program Files\\Common Files",
    "CommonFilesFolder": "C:\\Program Files (x86)\\Common Files",
    "DesktopFolder": "C:\\Users\\USER\\Desktop",
    "FavoritesFolder": "C:\\Users\\USER\\Favorites",
    "FontsFolder": "C:\\Windows\\Fonts",
    "LocalAppDataFolder": "C:\\Users\\USER\\AppData\\Local",
    "MyPicturesFolder": "C:\\Users\\USER\\Pictures",
    "NetHoodFolder": "C:\\Users\\USER\\AppData\\Roaming\\Microsoft\\Windows\\Network Shortcuts",
    "PersonalFolder": "C:\\Users\\USER\\Documents",
    "PrintHoodFolder": "C:\\Users\\USER\\AppData\\Roaming\\Microsoft\\Windows\\Printer Shortcuts",
    "ProgramFiles64Folder": "C:\\Program Files",
    "ProgramFilesFolder": "C:\\Program Files (x86)",
    "ProgramMenuFolder": "C:\\Users\\USER\\AppData\\Roaming\\Microsoft\\Windows\\Start Menu\\Programs",
    "RecentFolder": "C:\\Users\\USER\\AppData\\Roaming\\Microsoft\\Windows\\Recent",
    "SendToFolder": "C:\\Users\\USER\\AppData\\Roaming\\Microsoft\\Windows\\SendTo",
    "StartMenuFolder": "C:\\Users\\USER\\AppData\\Roaming\\Microsoft\\Windows\\Start Menu",
    "StartupFolder": "C:\\Users\\USER\\AppData\\Roaming\\Microsoft\\Windows\\Start Menu\\Programs\\Startup",
    "System16Folder": "C:\\Windows\\System",
    "System64Folder": "C:\\Windows\\System32",
    "SystemFolder": "C:\\Windows\\System32",
    "TempFolder": "C:\\Users\\USER\\AppData\\Local\\Temp",
    "TemplateFolder": "C:\\Users\\USER\\AppData\\Roaming\\Microsoft\\Windows\\Templates",
    "WindowsFolder": "C:\\Windows",
    "WindowsVolume": "C:\\",
}


def replace_root_id(system_folder_id: str) -> Optional[str]:
    path = ConfigManager().get(
        "ole", f"replacement_{system_folder_id}", DEFAULT_PATHS.get(system_folder_id)
    )
    if path:
        path = path.replace("\\", "/")
    return path


def supports_file(filetype: List[str]) -> bool:
    return "OLE" in filetype


# pylint: disable=too-many-positional-arguments
@surfactant.plugin.hookimpl
def extract_file_info(
    sbom: SBOM,
    software: Software,
    filename: str,
    filetype: List[str],
    software_field_hints: List[Tuple[str, object, int]],
    context_queue: "Queue[ContextEntry]",
    current_context: Optional[ContextEntry],
) -> object:
    if not supports_file(filetype):
        return None
    ole_info = extract_ole_info(filename)
    if ole_info and "ole" in ole_info:
        if "subject" in ole_info["ole"]:
            software_field_hints.append(("name", ole_info["ole"]["subject"], 80))
        if "revision_number" in ole_info["ole"]:
            software_field_hints.append(("version", ole_info["ole"]["revision_number"], 80))
        if "author" in ole_info["ole"]:
            software_field_hints.append(("vendor", ole_info["ole"]["author"], 80))
        if "comments" in ole_info["ole"]:
            software_field_hints.append(("comments", ole_info["ole"]["comments"], 80))

        if ole_info["ole"].get("clsid_type") == "MSI":
            file_decompression.create_extraction(
                filename, software, context_queue, current_context, extract_msi
            )

    return ole_info


def extract_ole_info(filename: str) -> Dict[str, Any]:
    file_details: Dict[str, Any] = {}

    with olefile.OleFileIO(filename) as ole:
        md = ole.get_metadata()
        file_details["ole"] = {}

        # to check if an OLE is an MSI file, check the root storage object CLSID
        # {000c1084-0000-0000-c000-000000000046}	MSI
        # {000c1086-0000-0000-c000-000000000046}    Windows Installer Patch MSP
        # extensions are typically .msi and .msp for files with these two clsid's
        # less common would be a .msm (merge) with the same clsid as MSI
        # as well as .mst (transform) with a clsid of 000c1082
        if ole.root and hasattr(ole.root, "clsid"):
            file_details["ole"]["clsid"] = str(ole.root.clsid).lower()
            if file_details["ole"]["clsid"] == "000c1082-0000-0000-c000-000000000046":
                file_details["ole"]["clsid_type"] = "MST"
            if file_details["ole"]["clsid"] == "000c1084-0000-0000-c000-000000000046":
                file_details["ole"]["clsid_type"] = "MSI"  # or msm, depending on file extension
            if file_details["ole"]["clsid"] == "000c1086-0000-0000-c000-000000000046":
                file_details["ole"]["clsid_type"] = "MSP"

        for prop in md.SUMMARY_ATTRIBS:
            if value := getattr(md, prop, None):
                if isinstance(value, bytes):
                    file_details["ole"][prop] = value.decode("unicode_escape")
                else:
                    file_details["ole"][prop] = str(value)

    return file_details


def extract_msi(filename: str, output_folder: str) -> List[Tuple[str, str]]:
    output_path = Path(output_folder)

    with pymsi.Package(Path(filename)) as package:
        msi = pymsi.Msi(package, True)

        preprocess_msi_decompression(msi)

        entries = extract_msi_root(msi.root, output_path)
        if entries:
            logger.debug(f"Extracted {entries}")
            logger.info(f"Extracted MSI contents to {output_path}")
        return entries


def preprocess_msi_decompression(msi: pymsi.Msi):
    folders: List[CabFolder] = []
    for media in msi.medias.values():
        if media.cabinet and media.cabinet.disks:
            for disk in media.cabinet.disks.values():
                for directory in disk:
                    for folder in directory.folders:
                        if folder not in folders:
                            folders.append(folder)

    total_folders = len(folders)
    logger.debug(f"Found {total_folders} folders in .cab files")

    # Use ThreadPoolExecutor to speed up decompression.
    # This is especially useful for LZX, which is implemented in pure Python.
    executor = ThreadPoolExecutor()
    completed_count = 0
    futures = {}
    try:
        for folder in folders:
            future = executor.submit(folder.decompress)
            futures[future] = folder

        for future in as_completed(futures):
            try:
                future.result()
                completed_count += 1
                folder = futures[future]
                logger.trace(f"Decompressed .cab folder: {folder}")
            except KeyboardInterrupt as e:
                raise e
            except (ValueError, RuntimeError, NotImplementedError, EOFError):
                logger.opt(exception=True).error(f"Error decompressing folder {futures[future]}")
    finally:
        for future in futures:
            future.cancel()
        executor.shutdown(wait=False)

    logger.debug("Decompression of .cab folders completed")


def extract_msi_root(root: Directory, output: Path) -> List[Tuple[str, str]]:
    if not output.exists():
        output.mkdir(parents=True, exist_ok=True)

    temp_installdir: Optional[Directory] = None

    def move_to_temp_installdir(item: Union[Component, Directory]):
        nonlocal temp_installdir
        if not temp_installdir:
            temp_installdir = Directory(
                {
                    "Directory": "_TEMPINSTALLDIR",
                    "Directory_Parent": root.id,
                    "DefaultDir": "INSTALLDIR",
                }
            )
            temp_installdir.parent = root
        if isinstance(item, Component):
            temp_installdir.components[item.id] = component
        elif isinstance(item, Directory):
            temp_installdir.children[item.id] = item
        else:
            raise TypeError(f"Unsupported item type: {type(item)}")

    for component in root.components.values():
        move_to_temp_installdir(component)

    entries = []

    for child in root.children.values():
        folder_name = child.id
        if "." in folder_name:
            folder_name, guid = folder_name.split(".", 1)
            if child.id != folder_name:
                logger.warning(f"MSI Directory ID '{child.id}' has a GUID suffix ({guid}).")

        if not replace_root_id(folder_name):
            logger.warning(
                f"MSI Directory ID '{folder_name}' has no replacement path defined. The MSI file may still be valid, but the path will not be accurate."
            )
            move_to_temp_installdir(child)
            continue
        extract_msi_directory(child, output / folder_name)
        entries.append((replace_root_id(folder_name), str(output / folder_name)))

    if temp_installdir is not None:
        extract_msi_directory(temp_installdir, output / temp_installdir.name)
        entries.append((None, str(output / temp_installdir.name)))

    return entries


def extract_msi_directory(root: Directory, output: Path):
    if not output.exists():
        output.mkdir(parents=True, exist_ok=True)

    for component in root.components.values():
        for file in component.files.values():
            if file.media is None:
                continue
            cab_file = file.resolve()
            (output / file.name).write_bytes(cab_file.decompress())

    for child in root.children.values():
        extract_msi_directory(child, output / child.name)



# Copyright 2023 Lawrence Livermore National Security, LLC
# See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: MIT
import gzip
import json
import subprocess
import tempfile
from typing import List, Optional

from loguru import logger

import surfactant.plugin
from surfactant.configmanager import ConfigManager
from surfactant.sbomtypes import SBOM, Software


class DockerScoutManager:
    def __init__(self) -> None:
        # Initialize ConfigManager
        config_manager = ConfigManager()

        # Retrieve the configuration option
        enable_docker_scout = config_manager.get("docker", "enable_docker_scout", True)

        # Set disable_docker_scout based on the configuration
        self.disable_docker_scout = not enable_docker_scout
        self.docker_scout_installed = False

    def check_docker_scout_installed(self) -> None:
        """Check if Docker Scout is installed and update the state accordingly."""
        if self.disable_docker_scout:
            return  # Do nothing if Docker Scout is disabled by config

        try:
            result = subprocess.run(["docker", "scout"], capture_output=True, check=False)
            self.docker_scout_installed = result.returncode == 0
        except FileNotFoundError:
            self.docker_scout_installed = False

        self.disable_docker_scout = not self.docker_scout_installed
        if not self.docker_scout_installed:
            logger.warning(
                "Install Docker Scout to scan containers for additional information. "
                "You can also disable this check by running 'surfactant config docker.enable_docker_scout false'."
            )

    def run_docker_scout(self, filename: str) -> object:
        """Run Docker Scout on the given file and return the results."""
        if self.disable_docker_scout:
            return {}  # Do nothing if Docker Scout is disabled by config

        try:
            result = subprocess.run(
                ["docker", "scout", "sbom", "--format", "spdx", f"fs://{filename}"],
                capture_output=True,
                check=False,
            )
            if result.returncode != 0:
                logger.warning(f"Running Docker Scout on {filename} failed")
                return {}
            spdx_out = json.loads(result.stdout)
            return {"dockerSPDX": spdx_out}
        except json.JSONDecodeError as e:
            logger.error(f"Failed to parse Docker Scout output for {filename}: {e}")
            return {}


# Initialize DockerScoutManager to check installation status
dsManager = DockerScoutManager()


def supports_file(filetype: List[str]) -> bool:
    """Check if the file type is supported."""
    if "DOCKER_TAR" in filetype or "DOCKER_GZIP" in filetype:
        return True
    return False


@surfactant.plugin.hookimpl
def extract_file_info(sbom: SBOM, software: Software, filename: str, filetype: List[str]) -> object:
    """Extract file information using Docker Scout if supported."""
    if dsManager.disable_docker_scout or not supports_file(filetype):
        return None
    return extract_docker_info(filetype, filename)


def extract_docker_info(filetype: List[str], filename: str) -> object:
    """Extract Docker information based on file type."""
    if "DOCKER_GZIP" in filetype:
        with open(filename, "rb") as gzip_in:
            gzip_data = gzip_in.read()
        with tempfile.NamedTemporaryFile() as gzip_out:
            gzip_out.write(gzip.decompress(gzip_data))
            gzip_out.flush()  # Ensure data is written before reading
            return dsManager.run_docker_scout(gzip_out.name)
    return dsManager.run_docker_scout(filename)


@surfactant.plugin.hookimpl
def init_hook(command_name: Optional[str] = None) -> None:
    if command_name != "update-db" and not dsManager.disable_docker_scout:
        dsManager.check_docker_scout_installed()
